package com.ijchart.xychart.plot ;

import java.awt.BasicStroke ;
import java.awt.Color ;
import java.awt.Graphics2D ;
import java.awt.Paint ;
import java.awt.Stroke ;
import java.awt.geom.Point2D ;
import java.awt.geom.Rectangle2D ;
import java.io.IOException ;
import java.io.ObjectInputStream ;
import java.io.ObjectOutputStream ;
import java.io.Serializable ;
import java.util.ArrayList ;
import java.util.Collection ;
import java.util.Collections ;
import java.util.List ;

import com.ijchart.xychart.axis.AbstractAxis ;
import com.ijchart.xychart.axis.AxisSpace ;
import com.ijchart.xychart.axis.IAxis ;
import com.ijchart.xychart.axis.category.CategoryAxis3D ;
import com.ijchart.xychart.axis.value.AbstractValueAxis ;
import com.ijchart.xychart.axis.value.NumberValueAxis ;
import com.ijchart.xychart.axis.value.NumberValueAxis3D ;
import com.ijchart.xychart.chartinfo.AnnotationShapeCollection ;
import com.ijchart.xychart.chartinfo.AxisShapeCollection ;
import com.ijchart.xychart.chartinfo.ChartShapeCollection ;
import com.ijchart.xychart.chartinfo.PlotShapeCollection ;
import com.ijchart.xychart.chartinfo.RenderShapeCollection ;
import com.ijchart.xychart.constant.AxisLocation ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.constant.PlotType ;
import com.ijchart.xychart.constant.RenderDrawOrder ;
import com.ijchart.xychart.data.XYNumberChartData ;
import com.ijchart.xychart.range.NumberRange ;
import com.ijchart.xychart.render.xy.AbstractXYRender ;
import com.ijchart.xychart.shape.LineChartShape ;
import com.ijchart.xychart.shape.RectangleTextChartShape ;
import com.ijchart.xychart.tickunit.INumberTickUnit ;
import com.ijchart.xychart.tickunit.LogTickUnit ;
import com.ijchart.xychart.ui.IEffect3D ;
import com.ijchart.xychart.ui.RectangleEdge ;
import com.ijchart.xychart.ui.RectangleInsets ;
import com.ijchart.xychart.ui.supplier.IDrawingSupplier ;
import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;
import com.ijchart.xychart.util.SerialUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ʾᶼ{@link com.ijchart.xychart.axis.AbstractAxis}plot.</zh>
 * <en>Used to display the two coordinate axis are
 * (@Link com.ijchart.xychart.axis.value.AbstractValueAxis) of the plot.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class XYPlot extends AbstractAxisPlot implements Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 3226938225090799837L ;

  /**
   * <zh>ȱʡvalueֵ᷽Ϊ0ĻǷɼ.</zh>
   * <en>A default flag that controls whether or not the zero baseline against the value
   * axis is visible.</en>
   */
  public static final boolean DEFAULT_VALUE_ZERO_BASELINE_VISIBLE = false ;

  /**
   * <zh>ȱʡԭ.</zh>
   * <en>The default origin point for the quadrants (if drawn).</en>
   */
  public static final XYNumberChartData DEFAULT_QUADRANT_ORIGIN
    = new XYNumberChartData(Double.NaN, Double.NaN) ;

  /**
   * <zh>ȱʡĻÿһ޵Paint.</zh>
   * <en>The default paint used for each quadrant. </en>
   */
  public static final Paint[] DEFAULT_QUADRANT_PAINT = new Paint[] {null, null, null, null} ;

  /**
   * <zh>render.</zh>
   * <en>Storage for the renders.</en>
   */
  private List renders ;

  /**
   * <zh>ԭ.</zh>
   * <en>The origin point for the quadrants (if drawn).</en>
   */
  private XYNumberChartData quadrantOrigin ;

  /**
   * <zh>ÿһ޵Paint.</zh>
   * <en>The paint used for each quadrant. </en>
   */
  private transient Paint[] quadrantPaint ;

  /**
   * <zh>ڴ洢domainλõļ.</zh>
   * <en>Storage for the domain axis locations.</en>
   */
  private List domainAxisLocations ;

  /**
   * <zh>domainֵ᷽Ϊ0ĻǷɼ.</zh>
   * <en>A flag that controls whether or not the zero baseline against the domain
   * axis is visible.</en>
   */
  private boolean domainZeroBaselineVisible ;

  /**
   * <zh>domainֵ᷽Ϊ0ĻߵStroke.</zh>
   * <en>The stroke used for the zero baseline against the domain axis.</en>
   */
  private transient Stroke domainZeroBaselineStroke ;

  /**
   * <zh>domainֵ᷽Ϊ0ĻߵPaint.</zh>
   * <en>The paint used for the zero baseline against the domain axis.</en>
   */
  private transient Paint domainZeroBaselinePaint ;

  /**
   * <zh>valueֵ᷽Ϊ0ĻǷɼ.</zh>
   * <en>A flag that controls whether or not the zero baseline against the value
   * axis is visible.</en>
   */
  private boolean valueZeroBaselineVisible ;

  /**
   * <zh>valueֵ᷽Ϊ0ĻߵStroke.</zh>
   * <en>The stroke used for the zero baseline against the value axis.</en>
   */
  private transient Stroke valueZeroBaselineStroke ;

  /**
   * <zh>valueֵ᷽Ϊ0ĻߵPaint.</zh>
   * <en>The paint used for the zero baseline against the value axis.</en>
   */
  private transient Paint valueZeroBaselinePaint ;

  /**
   * <zh>ȱʡĹ.</zh>
   * <en>Creates a new XYPlot instance with no dataset, no axes and
   * no render.  You should specify these items before using the plot.</en>
   */
  public XYPlot() {
    this(null, null, null) ;
  }

  /**
   * <zh>һµplot.</zh>
   * <en>Creates a new plot with the specified dataset, axes and render.  Any of the arguments can
   *  be null, but in that case you should take care to specify the value before using the plot
   * (otherwise a NullPointerException maybe thrown).</en>
   *
   * @param domainAxis AbstractValueAxis <zh>domain (Ϊnull).</zh>
   * <en>The domain axis (null permitted).</en>
   * @param valueAxis AbstractValueAxis <zh>value (Ϊnull).</zh>
   * <en>The value axis (null permitted).</en>
   * @param render AbstractXYRender <zh> (Ϊnull).</zh>
   * <en>The item render (null permitted).</en>
   */
  public XYPlot(AbstractValueAxis domainAxis, AbstractValueAxis valueAxis,
                AbstractXYRender render) {
    this.domainAxisLocations = new ArrayList() ;
    setListObject(this.domainAxes, 0, domainAxis) ;
    if (domainAxis != null) {
      domainAxis.addChangeListener(this) ;
    }

    setListObject(this.valueAxes, 0, valueAxis) ;
    if (valueAxis != null) {
      valueAxis.addChangeListener(this) ;
    }

    setDomainAxisLocation(0, AxisLocation.BOTTOM_OR_LEFT) ;
    setValueAxisLocation(0, AxisLocation.BOTTOM_OR_LEFT) ;

    this.renders = new ArrayList() ;
    setListObject(renders, 0, render) ;
    if (render != null) {
      render.addChangeListener(this) ;
    }

    this.domainZeroBaselineVisible = DEFAULT_DOMAIN_ZERO_BASELINE_VISIBLE ;
    this.domainZeroBaselinePaint = Color.black ;
    this.domainZeroBaselineStroke = new BasicStroke(0.5f) ;

    this.valueZeroBaselineVisible = DEFAULT_VALUE_ZERO_BASELINE_VISIBLE ;
    this.valueZeroBaselinePaint = Color.black ;
    this.valueZeroBaselineStroke = new BasicStroke(0.5f) ;

    this.quadrantOrigin = DEFAULT_QUADRANT_ORIGIN ;
    this.quadrantPaint = DEFAULT_QUADRANT_PAINT ;

    mapValueAxisToRender(0, 0) ;
  }

  /**
   * <zh>ֵΪ0domainλ.</zh>
   * <en>Returns the domain axis location for the primary domain axis.</en>
   *
   * @return <zh>λ (Ϊnull).</zh>
   * <en>The location (never null).</en>
   */
  public AxisLocation getDomainAxisLocation() {
    return getDomainAxisLocation(0) ;
  }

  /**
   * <zh>ָdomainλ.</zh>
   * <en>Returns the location for a domain axis.</en>
   *
   * @param index <zh>ֵ,ڵ0.</zh>
   * <en>The axis index,must be greater than or equal to 0.</en>
   *
   * @return <zh>λ.</zh>
   * <en>The location.</en>
   */
  public AxisLocation getDomainAxisLocation(int index) {
    if (index < 0) {
      throw new IllegalArgumentException("Must be 'index'>=0.") ;
    }

    AxisLocation result = null ;
    if (index < this.domainAxisLocations.size()) {
      result = (AxisLocation)this.domainAxisLocations.get(index) ;
    }
    if (result == null) {
      result = AxisLocation.getOpposite(getDomainAxisLocation(0)) ;
    }

    return result ;
  }

  /**
   * <zh>domainλ.</zh>
   * <en>Sets the location of the domain axis and, if requested, sends a
   * {@link com.ijchart.xychart.event.PlotChangeEvent} to all registered listeners.</en>
   *
   * @param location <zh>λ (Ϊnull).</zh>
   * <en>The axis location (null not permitted).</en>
   */
  public void setDomainAxisLocation(AxisLocation location) {
    setDomainAxisLocation(0, location) ;
  }

  /**
   * <zh>Ϊindexdomainλ.</zh>
   * <en>Sets the location for a domain axis and sends a
   * {@link com.ijchart.xychart.event.PlotChangeEvent} to all registered listeners.</en>
   *
   * @param index <zh>,ڵ0.</zh>
   * <en>The axis index,must be greater than or equal to 0.</en>
   * @param location <zh>λ (Ϊnull).</zh>
   * <en>The location (null not permitted).</en>
   */
  public void setDomainAxisLocation(int index, AxisLocation location) {
    if (index < 0) {
      throw new IllegalArgumentException("Must be 'index'>=0") ;
    }
    if (location == null) {
      throw new IllegalArgumentException("Null 'location' argument.") ;
    }

    setListObject(domainAxisLocations, index, location) ;
    firePlotChanged() ;
  }

  /**
   * <zh>0domainı߽λ.</zh>
   * <en>Returns the domain axis edge.  This is derived from the axis location
   * and the chart orientation.</en>
   *
   * @return <zh>߽λ.</zh>
   * <en>The edge (never null).</en>
   */
  public RectangleEdge getDomainAxisEdge() {
    return getDomainAxisEdge(0) ;
  }

  /**
   * <zh>ָdomainıֵ߽.</zh>
   * <en>Returns the edge for a domain axis.</en>
   *
   * @param index <zh>ֵ,ڵ0.</zh>
   * <en>The axis index,must be greater than or equal to 0.</en>
   *
   * @return <zh>ֵ.</zh>
   * <en>The edge (never null).</en>
   */
  public RectangleEdge getDomainAxisEdge(int index) {
    AxisLocation location = getDomainAxisLocation(index) ;
    if (location == null) {
      location = AxisLocation.BOTTOM_OR_LEFT ; //default value
    }

    return AbstractPlot.resolveDomainAxisLocation(location, getOrientation()) ;
  }

  /**
   * <zh>ص0domain axis.</zh>
   * <en>Returns the domain axis for the plot.</en>
   *
   * @return <zh>0domain axis (Ϊnull).</zh>
   * <en>The domain axis (null permitted).</en>
   */
  public AbstractValueAxis getDomainAxis() {
    return getDomainAxis(0) ;
  }

  /**
   * <zh>domain axis.</zh>
   * <en>Returns a domain axis.</en>
   *
   * @param index <zh>ֵ.</zh>
   * <en>The axis index.</en>
   *
   * @return <zh>domain axis (Ϊnull).</zh>
   * <en>The axis (null possible).</en>
   */
  public AbstractValueAxis getDomainAxis(int index) {
    AbstractValueAxis result = null ;
    if (index < this.domainAxes.size()) {
      result = (AbstractValueAxis)this.domainAxes.get(index) ;
    }

    return result ;
  }

  /**
   * <zh>domain axis.</zh>
   * <en>Returns a domain axis.</en>
   *
   * @param renderIndex <zh>renderֵ.</zh>
   * <en>The index of render.</en>
   *
   * @return <zh>0domain axis (Ϊnull).</zh>
   * <en>The axis (null possible).</en>
   */
  public AbstractValueAxis getDomainAxisByRenderIndex(int renderIndex) {
    Integer axisIndex = (Integer)this.renderToDomainAxis.get(new Integer(renderIndex)) ;
    if (axisIndex != null) {
      int idx = axisIndex.intValue() ;
      if (idx < this.domainAxes.size()) {
        return (AbstractValueAxis)this.domainAxes.get(idx) ;
      } else {
        return (AbstractValueAxis)this.domainAxes.get(this.domainAxes.size() - 1) ;
      }
    }

    if (renderIndex < this.domainAxes.size()) {
      return (AbstractValueAxis)this.domainAxes.get(renderIndex) ;
    } else {
      return (AbstractValueAxis)this.domainAxes.get(this.domainAxes.size() - 1) ;
    }
  }

  /**
   * <zh>plotdomain,м
   * {@link com.ijchart.xychart.event.PlotChangeEvent}¼.</zh>
   * <en>Sets the domain axis for the plot and sends a
   *  {@link com.ijchart.xychart.event.PlotChangeEvent} to all registered listeners.</en>
   *
   * @param axis AbstractValueAxis <zh>domain (Ϊnull).</zh>
   * <en>The axis (null not permitted).</en>
   */
  public void setDomainAxis(AbstractValueAxis axis) {
    setDomainAxis(0, axis, true) ;
  }

  /**
   * <zh>domain,м{@link com.ijchart.xychart.event.PlotChangeEvent}¼.</zh>
   * <en>Sets a domain axis and sends a {@link com.ijchart.xychart.event.PlotChangeEvent}
   *  to all registered listeners.</en>
   *
   * @param index int <zh>ֵ,ڵ0.</zh>
   * <en>The axis index,must be greater than or equal to 0.</en>
   * @param axis AbstractValueAxis <zh>domain (Ϊnull).</zh>
   * <en>The axis (null not permitted).</en>
   */
  public void setDomainAxis(int index, AbstractValueAxis axis) {
    setDomainAxis(index, axis, true) ;
  }

  /**
   * <zh>domain,м{@link com.ijchart.xychart.event.PlotChangeEvent}¼.</zh>
   * <en>Sets a domain axis and sends a {@link com.ijchart.xychart.event.PlotChangeEvent}
   *  to all registered listeners.</en>
   *
   * @param index int <zh>ֵ,ڵ0.</zh>
   * <en>The axis index,must be greater than or equal to 0.</en>
   * @param axis AbstractValueAxis <zh>domain (Ϊnull).</zh>
   * <en>The axis (null not permitted).</en>
   * @param notify boolean <zh>Ƿ֪ͨ.</zh>
   * <en>Whether or not notify listener.</en>
   */
  public void setDomainAxis(int index, AbstractValueAxis axis, boolean notify) {
    if (index < 0) {
      throw new IllegalArgumentException("Must be 'index'>=0.") ;
    }
    if (axis == null) {
      throw new IllegalArgumentException("Null 'axis' argument.") ;
    }

    if (index < this.domainAxes.size()) {
      AbstractValueAxis existing = (AbstractValueAxis)this.domainAxes.get(index) ;
      if (existing != null) {
        existing.removeChangeListener(this) ;
      }
    }

    setListObject(domainAxes, index, axis) ;
    axis.addChangeListener(this) ;

    if (notify) {
      firePlotChanged() ;
    }
  }

  /**
   * <zh>plotdomain,ѾڵĶᱻ,м
   * {@link com.ijchart.xychart.event.PlotChangeEvent}¼.</zh>
   * <en>Sets the domain axes for this plot and sends a
   *  {@link com.ijchart.xychart.event.PlotChangeEvent} to all registered listeners.</en>
   *
   * @param axes AbstractValueAxis[] <zh>domain (Ϊnull).</zh>
   * <en>The axes (null not permitted).</en>
   */
  public void setDomainAxes(AbstractValueAxis[] axes) {
    setDomainAxes(axes, 0) ;
  }

  /**
   * <zh>plotdomain,м
   * {@link com.ijchart.xychart.event.PlotChangeEvent}¼.</zh>
   * <en>Sets the domain axes for this plot and sends a
   *  {@link com.ijchart.xychart.event.PlotChangeEvent} to all registered listeners.</en>
   *
   * @param axes AbstractValueAxis[] <zh>domain (Ϊnull).</zh>
   * <en>The axes (null not permitted).</en>
   * @param startIndex int <zh>Ŀʼֵ,startIndex֮ǰĶ󲻻ᱻ,ڵ0.</zh>
   * <en>The beginning of the index value, startIndex before the target will not be covered.
   * must be greater than or equal to 0.</en>
   */
  public void setDomainAxes(AbstractValueAxis[] axes, int startIndex) {
    if (axes == null) {
      throw new IllegalArgumentException("Null 'axes' argument.") ;
    }
    if (startIndex < 0) {
      throw new IllegalArgumentException("Must be 'startIndex'>=0.") ;
    }

    if (axes.length > 0) {
      for (int i = 0 ; i < axes.length ; i++) {
        setDomainAxis(startIndex + i, axes[i], false) ;
      }
      firePlotChanged() ;
    }
  }

  /**
   * <zh>domainĸ.</zh>
   * <en>Returns the number of category axes.</en>
   *
   * @return <zh>ĸ.</zh>
   * <en>The axis count.</en>
   */
  public int getDomainAxisCount() {
    return this.domainAxes.size() ;
  }

  /**
   * <zh>ָdomain.</zh>
   * <en>Returns the index of the specified axis, or -1 if the axis
   * is not assigned to the plot.</en>
   *
   * @param axis <zh>domain (Ϊnull).</zh>
   * <en>The axis (null not permitted).</en>
   *
   * @return <zh>ֵ,ûҵ򷵻-1.</zh>
   * <en>The axis index.</en>
   */
  public int getDomainAxisIndex(AbstractValueAxis axis) {
    if (axis == null) {
      throw new IllegalArgumentException("Null 'axis' argument.") ;
    }

    return this.domainAxes.indexOf(axis) ;
  }

  /**
   * <zh>domainֵ᷽Ϊ0ĻǷɼ.</zh>
   * <en>Returns a flag that controls whether or not a zero baseline is
   * displayed for the domain axis.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isDomainZeroBaselineVisible() {
    return this.domainZeroBaselineVisible ;
  }

  /**
   * <zh>domainֵ᷽Ϊ0ĻǷɼ,еļ
   * {@link com.ijchart.xychart.event.PlotChangeEvent}¼.</zh>
   * <en>Sets the flag that controls whether or not the zero baseline is displayed for the domain
   * axis, and sends a {@link com.ijchart.xychart.event.PlotChangeEvent}
   *  to all registered listeners.</en>
   *
   * @param visible boolean <zh>ֵ.</zh>
   * <en>The flag.</en>
   */
  public void setDomainZeroBaselineVisible(boolean visible) {
    if (this.domainZeroBaselineVisible != visible) {
      this.domainZeroBaselineVisible = visible ;
      firePlotChanged() ;
    }
  }

  /**
   * <zh>ػdomainֵ᷽Ϊ0ĻߵStroke.</zh>
   * <en>Returns the stroke used for the zero baseline against the domain axis.</en>
   *
   * @return Stroke <zh>Stroke (Ϊnull)</zh>
   * <en>The stroke (never null).</en>
   */
  public Stroke getDomainZeroBaselineStroke() {
    return this.domainZeroBaselineStroke ;
  }

  /**
   * <zh>ûdomainֵ᷽Ϊ0ĻߵStroke,еļ
   * {@link com.ijchart.xychart.event.PlotChangeEvent}¼.</zh>
   * <en>Sets the stroke for the zero baseline for the domain axis,
   * and sends a {@link com.ijchart.xychart.event.PlotChangeEvent} to all registered listeners.</en>
   *
   * @param stroke Stroke <zh>Stroke (Ϊnull).</zh>
   * <en>The stroke (null not permitted).</en>
   */
  public void setDomainZeroBaselineStroke(Stroke stroke) {
    if (stroke == null) {
      throw new IllegalArgumentException("Null 'stroke' argument.") ;
    }
    this.domainZeroBaselineStroke = stroke ;
    firePlotChanged() ;
  }

  /**
   * <zh>ػdomainֵ᷽Ϊ0ĻߵPaint.</zh>
   * <en>Returns the paint for the zero baseline (if any) plotted against the domain axis.</en>
   *
   * @return Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (never null).</en>
   */
  public Paint getDomainZeroBaselinePaint() {
    return this.domainZeroBaselinePaint ;
  }

  /**
   * <zh>ûdomainֵ᷽Ϊ0ĻߵPaint,еļ
   * {@link com.ijchart.xychart.event.PlotChangeEvent}¼.</zh>
   * <en>Sets the paint for the zero baseline plotted against the domain axis and
   * sends a {@link com.ijchart.xychart.event.PlotChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (null not permitted).</en>
   */
  public void setDomainZeroBaselinePaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument.") ;
    }
    this.domainZeroBaselinePaint = paint ;
    firePlotChanged() ;
  }

  /**
   * <zh>valueֵ᷽Ϊ0ĻǷɼ.</zh>
   * <en>Returns a flag that controls whether or not a zero baseline is
   * displayed for the value axis.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isValueZeroBaselineVisible() {
    return this.valueZeroBaselineVisible ;
  }

  /**
   * <zh>valueֵ᷽Ϊ0ĻǷɼ,еļ
   * {@link com.ijchart.xychart.event.PlotChangeEvent}¼.</zh>
   * <en>Sets the flag that controls whether or not the zero baseline is displayed for the value
   * axis, and sends a {@link com.ijchart.xychart.event.PlotChangeEvent}
   *  to all registered listeners.</en>
   *
   * @param visible boolean <zh>ֵ.</zh>
   * <en>The flag.</en>
   */
  public void setValueZeroBaselineVisible(boolean visible) {
    if (this.valueZeroBaselineVisible != visible) {
      this.valueZeroBaselineVisible = visible ;
      firePlotChanged() ;
    }
  }

  /**
   * <zh>ػvalueֵ᷽Ϊ0ĻߵStroke.</zh>
   * <en>Returns the stroke used for the zero baseline against the value axis.</en>
   *
   * @return Stroke <zh>Stroke (Ϊnull)</zh>
   * <en>The stroke (never null).</en>
   */
  public Stroke getValueZeroBaselineStroke() {
    return this.valueZeroBaselineStroke ;
  }

  /**
   * <zh>ûvalueֵ᷽Ϊ0ĻߵStroke,еļ
   * {@link com.ijchart.xychart.event.PlotChangeEvent}¼.</zh>
   * <en>Sets the stroke for the zero baseline for the value axis,
   * and sends a {@link com.ijchart.xychart.event.PlotChangeEvent} to all registered listeners.</en>
   *
   * @param stroke Stroke <zh>Stroke (Ϊnull).</zh>
   * <en>The stroke (null not permitted).</en>
   */
  public void setValueZeroBaselineStroke(Stroke stroke) {
    if (stroke == null) {
      throw new IllegalArgumentException("Null 'stroke' argument.") ;
    }
    this.valueZeroBaselineStroke = stroke ;
    firePlotChanged() ;
  }

  /**
   * <zh>ػvalueֵ᷽Ϊ0ĻߵPaint.</zh>
   * <en>Returns the paint for the zero baseline (if any) plotted against the value axis.</en>
   *
   * @return Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (never null).</en>
   */
  public Paint getValueZeroBaselinePaint() {
    return this.valueZeroBaselinePaint ;
  }

  /**
   * <zh>ûvalueֵ᷽Ϊ0ĻߵPaint,еļ
   * {@link com.ijchart.xychart.event.PlotChangeEvent}¼.</zh>
   * <en>Sets the paint for the zero baseline plotted against the value axis and
   * sends a {@link com.ijchart.xychart.event.PlotChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (null not permitted).</en>
   */
  public void setValueZeroBaselinePaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument.") ;
    }
    this.valueZeroBaselinePaint = paint ;
    firePlotChanged() ;
  }

  /**
   * <zh>ԭ.</zh>
   * <en>Returns the origin point for the quadrants (if drawn).</en>
   *
   * @return Point2D <zh>Point2D (Ϊnull).</zh>
   * <en>The origin point (never null).</en>
   */
  public XYNumberChartData getQuadrantOrigin() {
    return this.quadrantOrigin ;
  }

  /**
   * <zh>ԭ,еļ{@link com.ijchart.xychart.event.PlotChangeEvent}¼.</zh>
   * <en>Sets the quadrant origin and sends a {@link com.ijchart.xychart.event.PlotChangeEvent}
   * to all registered listeners.</en>
   *
   * @param origin XYNumberChartData <zh>XYNumberChartData (Ϊnull).</zh>
   * <en>The origin point (null not permitted).</en>
   */
  public void setQuadrantOrigin(XYNumberChartData origin) {
    if (origin == null) {
      throw new IllegalArgumentException("Null 'origin' argument.") ;
    }

    this.quadrantOrigin = origin ;
    firePlotChanged() ;
  }

  /**
   * <zh>ػÿһ޵Paint.</zh>
   * <en>Returns the paint used for the specified quadrant.</en>
   *
   * @param index int <zh>ֵ [0-3].</zh>
   * <en>The quadrant index (0-3).</en>
   *
   * @return Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (maybe null).</en>
   */
  public Paint getQuadrantPaint(int index) {
    if (index < 0 || index > 3) {
      throw new IllegalArgumentException("The index value (" + index
                                         + ") should be in the value 0 to 3.") ;
    }

    return this.quadrantPaint[index] ;
  }

  /**
   * <zh>ûÿһ޵Paint,еļ
   * {@link com.ijchart.xychart.event.PlotChangeEvent}¼.</zh>
   * <en>Sets the paint used for the specified quadrant and sends a
   * {@link com.ijchart.xychart.event.PlotChangeEvent} to all registered listeners.</en>
   *
   * @param index int <zh>ֵ [0-3].</zh>
   * <en>The quadrant index (0-3).</en>
   * @param paint Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (null permitted).</en>
   */
  public void setQuadrantPaint(int index, Paint paint) {
    if (index < 0 || index > 3) {
      throw new IllegalArgumentException("The index value (" + index
                                         + ") should be in the range 0 to 3.") ;
    }

    this.quadrantPaint[index] = paint ;
    firePlotChanged() ;
  }

  /**
   * <zh>domain.</zh>
   * <en>Zoom domain axes.</en>
   *
   * @param selectArea Rectangle2D <zh>ѡ,Ϊnullʾָǰ״̬.</zh>
   * <en>The selected area,null to restore chart.</en>
   * @param dataArea Rectangle2D <zh>,Ϊnullʾָǰ״̬.</zh>
   * <en>The data area,null to restore chart.</en>
   * @param source Point2D <zh>Դ (Ϊnull).</zh>
   * <en>The source mouse point (null not permitted).</en>
   */
  public void zoomDomainAxes(Rectangle2D selectArea, Rectangle2D dataArea, Point2D source) {
    if (source == null) {
      throw new IllegalArgumentException("Null 'source' argument.") ;
    }

    boolean restore = (selectArea == null || dataArea == null) ;
    for (int i = 0 ; i < this.domainAxes.size() ; i++) {
      AbstractAxis valueAxis = (AbstractAxis)this.domainAxes.get(i) ;
      if (valueAxis != null && valueAxis instanceof NumberValueAxis) {
        NumberValueAxis numberAxis = (NumberValueAxis) valueAxis ;
        if (restore) {
          numberAxis.setZoomDatasetRange(null) ;
          continue ;
        }

        NumberRange datasetRange = numberAxis.getTickRange() ;
        if (datasetRange == null) {
          continue ;
        }

        double lowerPercent = 0, upperPercent = 0 ;
        if (getOrientation().equals(ChartOrientation.VERTICAL)) {
          double lowerMarginLen = numberAxis.getLowerMargin() * dataArea.getWidth() ;
          double upperMarginLen = numberAxis.getUpperMargin() * dataArea.getWidth() ;

          double lowerDataAreaX = dataArea.getX() + lowerMarginLen ;
          double upperDataAreaX = dataArea.getMaxX() - upperMarginLen ;

          double selectAreaMinX = Math.max(selectArea.getMinX(), lowerDataAreaX) ;
          double selectAreaMaxX = Math.min(selectArea.getMaxX(), upperDataAreaX) ;
          double dataAreaDrawWidth = dataArea.getWidth() - lowerMarginLen - upperMarginLen ;

          lowerPercent = Math.abs(selectAreaMinX - lowerDataAreaX) / dataAreaDrawWidth ;
          upperPercent = Math.abs(selectAreaMaxX - upperDataAreaX) / dataAreaDrawWidth ;

        } else if (getOrientation().equals(ChartOrientation.HORIZONTAL)) {
          double lowerMarginLen = numberAxis.getLowerMargin() * dataArea.getHeight() ;
          double upperMarginLen = numberAxis.getUpperMargin() * dataArea.getHeight() ;

          double upperDataAreaY = dataArea.getY() + upperMarginLen ;
          double lowerDataAreaY = dataArea.getMaxY() - lowerMarginLen ;

          double selectAreaMinY = Math.max(selectArea.getMinY(), upperDataAreaY) ;
          double selectAreaMaxY = Math.min(selectArea.getMaxY(), lowerDataAreaY) ;
          double dataAreaDrawHeight = dataArea.getHeight() - lowerMarginLen - upperMarginLen ;

          lowerPercent = Math.abs(selectAreaMaxY - lowerDataAreaY) / dataAreaDrawHeight ;
          upperPercent = Math.abs(selectAreaMinY - upperDataAreaY) / dataAreaDrawHeight ;

          NumberRange zoomRange = datasetRange.zoomRange(lowerPercent, upperPercent,
                                                         numberAxis.isInverted()) ;
          numberAxis.setZoomDatasetRange(zoomRange) ;
        }

        INumberTickUnit tickUnit = numberAxis.getTickUnit() ;
        if (tickUnit instanceof LogTickUnit) {
          LogTickUnit logTickUnit = (LogTickUnit) tickUnit ;
          String[] labels = logTickUnit.getExponentLabels() ;
          NumberRange range = new NumberRange(Double.parseDouble(labels[0]),
                                              Double.parseDouble(labels[labels.length - 1])) ;
          NumberRange tmpZoomRange = range.zoomRange(lowerPercent, upperPercent,
                                                     numberAxis.isInverted()) ;
          double rangeLower = Math.pow(logTickUnit.getLogarithmBase(), tmpZoomRange.getLowerBound()) ;
          double rangeUpper = Math.pow(logTickUnit.getLogarithmBase(), tmpZoomRange.getUpperBound()) ;
          NumberRange zoomRange = new NumberRange(rangeLower, rangeUpper) ;
          numberAxis.setZoomDatasetRange(zoomRange) ;
        } else {
          NumberRange zoomRange = datasetRange.zoomRange(lowerPercent, upperPercent,
                                                         numberAxis.isInverted()) ;
          numberAxis.setZoomDatasetRange(zoomRange) ;
        }

      }
    }
  }

  /**
   * <zh>ͼplot,Ҫʵִ˷.</zh>
   * <en>Draws the plot within the specified area.Subclasses need to provide an
   * implementation of this method, obviously.</en>
   *
   * @param g2 Graphics2D<zh>ͼ豸(ӡĻ) (Ϊnull).</zh>
   * <en>The graphics device (null not permitted).</en>
   * @param plotArea Rectangle2D <zh>plot (Ϊnull).</zh>
   * <en>The plot area (null not permitted).</en>
   * @param chartShapes ChartShapeCollection <zh>chartͼμ (Ϊnull).</zh>
   * <en>Collects information about the chart (null not permitted).</en>
   */
  public void draw(Graphics2D g2, Rectangle2D plotArea, ChartShapeCollection chartShapes) {
    if (g2 == null) {
      throw new IllegalArgumentException("Null 'g2' argument.") ;
    }
    if (plotArea == null) {
      throw new IllegalArgumentException("Null 'plotArea' argument.") ;
    }
    if (chartShapes == null) {
      throw new IllegalArgumentException("Null 'chartShapes' argument.") ;
    }

    // if the plot area is too small, just return...
    if (plotArea.getWidth() <= DEFAULT_MINIMUM_WIDTH_TO_DRAW ||
        plotArea.getHeight() <= DEFAULT_MINIMUM_HEIGHT_TO_DRAW) {
      return ;
    }

    plotShapes = new PlotShapeCollection() ;
    plotShapes.setDrawArea(plotArea) ;
    chartShapes.addPlotShapes(plotShapes) ;

    ChartOrientation orientation = getOrientation() ;

    // adjust the drawing area for the plot insets (if any)...
    RectangleInsets insets = getInsets() ;
    insets.trim(plotArea) ;

    drawBorder(plotArea, plotShapes) ;
    drawBackground(plotArea, orientation, plotShapes) ;
    drawBackgroundImage(plotArea, plotShapes) ;

    //еݼΪ,ʾݼʱϢ
    boolean isAllDatasetEmpty = true ;
    for (int i = 0 ; i < this.valueAxes.size() ; i++) {
      AbstractAxis valueAxis = (AbstractAxis)this.valueAxes.get(i) ;
      if (valueAxis != null) {
        if (!valueAxis.getDataset().isEmpty()) {
          isAllDatasetEmpty = false ;
          break ;
        }
      }
    }
    if (isAllDatasetEmpty) {
      drawNoDataMessage(plotArea, plotShapes) ;
      return ;
    }

    allocateRenderShapesList(this.renders.size()) ;
    RenderShapeCollection renderShapes = (RenderShapeCollection)this.renderShapesList.get(0) ;
    plotShapes.addRenderShape(renderShapes) ;

    this.dataArea = drawAxes(g2, plotArea, plotShapes) ;
    if (this.dataArea == null) {
      return ;
    }
    renderShapes.setDrawArea(this.dataArea) ;
    drawQuadrantBackground(this.dataArea, orientation, getDomainAxis(), getValueAxis(),
                           renderShapes) ;
    drawZeroDomainBaseline(this.dataArea, plotShapes) ;
    drawZeroValueBaseline(this.dataArea, plotShapes) ;

    //draw grid lines
    drawDomainGridlines(g2, this.dataArea, renderShapes) ;
    drawValueGridlines(g2, this.dataArea, renderShapes, this.valueAxes.size()) ;

    if (getRenderDrawOrder().equals(RenderDrawOrder.FORWARD)) {
      IDrawingSupplier supplier = getDrawingSupplier() ;
      supplier.reset() ;
      for (int i = 0 ; i < this.renders.size() ; i++) {
        if (this.renders.get(i) == null
            || !(this.renders.get(i) instanceof AbstractXYRender)) {
          continue ;
        }

        AbstractXYRender render = (AbstractXYRender)this.renders.get(i) ;
        render.setDrawingSupplier(supplier) ;
        if (i == 0) { //0renderƱͼƬ,Ĳ
          render.drawBackground(this.dataArea, orientation, renderShapes) ;
          render.drawDrawAreaBorder(this.dataArea, renderShapes) ;
          render.drawBackgroundImage(this.dataArea, renderShapes) ;
          AbstractValueAxis domainAxis = getDomainAxisByRenderIndex(i) ;
          RectangleEdge dEdge = getDomainAxisEdge(i) ;
          if (domainAxis != null) {
            render.drawValueMarker(domainAxis, false, dEdge, this.dataArea, orientation,
                                   renderShapes) ;
            render.drawIntervalValueMarker(domainAxis, false, dEdge, this.dataArea, orientation,
                                           renderShapes) ;
          }

          AbstractValueAxis valueAxis = getValueAxisByRenderIndex(i) ;
          RectangleEdge vEdge = getValueAxisEdge(i) ;
          if (valueAxis != null) {
            render.drawValueMarker(valueAxis, true, vEdge, this.dataArea, orientation, renderShapes) ;
            render.drawIntervalValueMarker(valueAxis, true, vEdge, this.dataArea, orientation,
                                           renderShapes) ;
          }

          if (domainAxis != null && valueAxis != null) {
            renderShapes.setDataset(render.getDataset()) ;
            render.drawAllItems(domainAxis, dEdge, valueAxis, vEdge, this.dataArea, orientation,
                                renderShapes, getForegroundAlpha()) ;

            AnnotationShapeCollection annotationShapes = new AnnotationShapeCollection() ;
            annotationShapes.setDrawArea(this.dataArea) ;
            plotShapes.addAnnotationShape(annotationShapes) ;
            render.drawAllAnnotation(domainAxis, dEdge, valueAxis, vEdge, this.dataArea,
                                     orientation, annotationShapes) ;
          }
        } else {
          RenderShapeCollection tmpRenderShapes = (RenderShapeCollection) renderShapesList.get(i) ;
          if (tmpRenderShapes == null) {
            return ;
          }
          tmpRenderShapes.setDrawArea(this.dataArea) ;
          plotShapes.addRenderShape(tmpRenderShapes) ;

          AbstractValueAxis domainAxis = getDomainAxisByRenderIndex(i) ;
          RectangleEdge dEdge = getDomainAxisEdge(i) ;
          if (domainAxis != null) {
            render.drawValueMarker(domainAxis, false, dEdge, this.dataArea, orientation,
                                   renderShapes) ;
            render.drawIntervalValueMarker(domainAxis, false, dEdge, this.dataArea, orientation,
                                           renderShapes) ;
          }

          AbstractValueAxis valueAxis = getValueAxisByRenderIndex(i) ;
          RectangleEdge vEdge = getValueAxisEdge(i) ;
          if (valueAxis != null) {
            render.drawValueMarker(valueAxis, true, vEdge, this.dataArea, orientation,
                                   tmpRenderShapes) ;
            render.drawIntervalValueMarker(valueAxis, true, vEdge, this.dataArea, orientation,
                                           tmpRenderShapes) ;
          }

          if (domainAxis == null) {
            domainAxis = getDomainAxis(0) ;
          }
          if (valueAxis == null) {
            valueAxis = getValueAxis(0) ;
          }

          if (domainAxis != null && valueAxis != null) {
            tmpRenderShapes.setDataset(render.getDataset()) ;
            render.drawAllItems(domainAxis, dEdge, valueAxis, vEdge, this.dataArea, orientation,
                                tmpRenderShapes, getForegroundAlpha()) ;

            AnnotationShapeCollection annotationShapes = new AnnotationShapeCollection() ;
            annotationShapes.setDrawArea(this.dataArea) ;
            plotShapes.addAnnotationShape(annotationShapes) ;
            render.drawAllAnnotation(domainAxis, dEdge, valueAxis, vEdge, this.dataArea,
                                     orientation, annotationShapes) ;

          }
        }

      }
    } else {
      IDrawingSupplier supplier = getDrawingSupplier() ;
      supplier.reset() ;

      for (int i = this.renders.size() - 1 ; i >= 0 ; i--) {
        if (this.renders.get(i) == null
            || !(this.renders.get(i) instanceof AbstractXYRender)) {
          continue ;
        }

        AbstractXYRender render = (AbstractXYRender)this.renders.get(i) ;
        render.setDrawingSupplier(supplier) ;
        if (i == this.renders.size() - 1) { //һrenderƱͼƬ,Ĳ
          render.drawBackground(this.dataArea, orientation, renderShapes) ;
          render.drawDrawAreaBorder(this.dataArea, renderShapes) ;
          render.drawBackgroundImage(this.dataArea, renderShapes) ;

          AbstractValueAxis domainAxis = getDomainAxisByRenderIndex(i) ;
          RectangleEdge dEdge = getDomainAxisEdge(i) ;
          if (domainAxis != null) {
            render.drawValueMarker(domainAxis, false, dEdge, this.dataArea, orientation,
                                   renderShapes) ;
            render.drawIntervalValueMarker(domainAxis, false, dEdge, this.dataArea, orientation,
                                           renderShapes) ;
          }

          AbstractValueAxis valueAxis = getValueAxisByRenderIndex(i) ;
          RectangleEdge vEdge = getValueAxisEdge(i) ;
          if (valueAxis != null) {
            render.drawValueMarker(valueAxis, true, vEdge, this.dataArea, orientation, renderShapes) ;
            render.drawIntervalValueMarker(valueAxis, true, vEdge, this.dataArea, orientation,
                                           renderShapes) ;
          }

          if (domainAxis == null) {
            domainAxis = getDomainAxis(0) ;
          }
          if (valueAxis == null) {
            valueAxis = getValueAxis(0) ;
          }

          if (domainAxis != null && valueAxis != null) {
            renderShapes.setDataset(render.getDataset()) ;
            render.drawAllItems(domainAxis, dEdge, valueAxis, vEdge, this.dataArea, orientation,
                                renderShapes, getForegroundAlpha()) ;

            AnnotationShapeCollection annotationShapes = new AnnotationShapeCollection() ;
            annotationShapes.setDrawArea(this.dataArea) ;
            plotShapes.addAnnotationShape(annotationShapes) ;
            render.drawAllAnnotation(domainAxis, dEdge, valueAxis, vEdge, this.dataArea,
                                     orientation, annotationShapes) ;
          }
        } else {
          RenderShapeCollection tmpRenderShapes = (RenderShapeCollection) renderShapesList.get(i) ;
          if (tmpRenderShapes == null) {
            return ;
          }
          tmpRenderShapes.setDrawArea(this.dataArea) ;
          plotShapes.addRenderShape(tmpRenderShapes) ;

          AbstractValueAxis domainAxis = getDomainAxisByRenderIndex(i) ;
          RectangleEdge dEdge = getDomainAxisEdge(i) ;
          if (domainAxis != null) {
            render.drawValueMarker(domainAxis, false, dEdge, dataArea, orientation, renderShapes) ;
            render.drawIntervalValueMarker(domainAxis, false, dEdge, this.dataArea, orientation,
                                           renderShapes) ;
          }

          AbstractValueAxis valueAxis = getValueAxisByRenderIndex(i) ;
          RectangleEdge vEdge = getValueAxisEdge(i) ;
          if (valueAxis != null) {
            render.drawValueMarker(valueAxis, true, vEdge, dataArea, orientation, tmpRenderShapes) ;
            render.drawIntervalValueMarker(valueAxis, true, vEdge, this.dataArea, orientation,
                                           tmpRenderShapes) ;
          }

          if (domainAxis == null) {
            domainAxis = getDomainAxis(0) ;
          }
          if (valueAxis == null) {
            valueAxis = getValueAxis(0) ;
          }

          if (domainAxis != null && valueAxis != null) {
            tmpRenderShapes.setDataset(render.getDataset()) ;
            render.drawAllItems(domainAxis, dEdge, valueAxis, vEdge, this.dataArea, orientation,
                                tmpRenderShapes, getForegroundAlpha()) ;

            AnnotationShapeCollection annotationShapes = new AnnotationShapeCollection() ;
            annotationShapes.setDrawArea(this.dataArea) ;
            plotShapes.addAnnotationShape(annotationShapes) ;
            render.drawAllAnnotation(domainAxis, dEdge, valueAxis, vEdge, this.dataArea,
                                     orientation, annotationShapes) ;
          }
        }
      }
    }

    translateMarkerShapes() ;
  }

  /**
   * <zh>е.</zh>
   * <en>A utility method for drawing the plot's axes.</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(ӡĻ) (Ϊnull).</zh>
   * <en>The graphics device (null not permitted).</en>
   * @param plotArea Rectangle2D <zh>plot (Ϊnull).</zh>
   * <en>The plot draw area (null not permitted).</en>
   * @param shapeCollection PlotShapeCollection <zh>plotͼμ (Ϊnull).</zh>
   * <en>Collects information about the plot (null not permitted).</en>
   *
   * @return Rectangle2D <zh>.</zh>
   * <en>The data area.</en>
   */
  protected Rectangle2D drawAxes(Graphics2D g2, Rectangle2D plotArea,
                                 PlotShapeCollection shapeCollection) {
    if (g2 == null) {
      throw new IllegalArgumentException("Null 'g2' argument.") ;
    }
    if (plotArea == null) {
      throw new IllegalArgumentException("Null 'plotArea' argument.") ;
    }
    if (shapeCollection == null) {
      throw new IllegalArgumentException("Null 'shapeCollection' argument.") ;
    }

    AxisSpace axisSpace = new AxisSpace() ; //Ԥȼspace
    Rectangle2D clonePlotArea = (Rectangle2D) plotArea.clone() ;

    int categoryAxesCount = 0 ;
    for (int i = 0 ; i < this.domainAxes.size() ; i++) {
      IAxis xAxis = (IAxis)this.domainAxes.get(i) ;
      if (xAxis != null) {
        categoryAxesCount++ ;
      }
    }

    //calculate data area size
    boolean firstLeftAxis = false ;
    double[] categoryAxisWidthArr = new double[this.domainAxes.size()] ;
    for (int i = 0 ; i < this.domainAxes.size() ; i++) {
      IAxis xAxis = (IAxis)this.domainAxes.get(i) ;
      if (xAxis != null) {
        RectangleEdge edge = getDomainAxisEdge(i) ;
        if (categoryAxesCount <= 1 && !firstLeftAxis
            && (xAxis instanceof IEffect3D) && edge.equals(RectangleEdge.LEFT)) { //3D͵category
          edge = RectangleEdge.RIGHT ;
          firstLeftAxis = true ;
        }
        double fixedSpace = xAxis.getFixedAxisSpace() ;
        double axisWidth = 0 ;
        if (fixedSpace > 0) {
          axisWidth = fixedSpace ;
        } else {
          axisWidth = xAxis.calculateAxisSpace(g2, plotArea, clonePlotArea, edge) ;
        }

        axisSpace.add(axisWidth, edge) ;
        categoryAxisWidthArr[i] = axisWidth ;
        edge.trim(clonePlotArea, axisWidth) ;
      }
    }

    double[] valueAxisWidthArr = new double[this.valueAxes.size()] ;
    for (int i = 0 ; i < this.valueAxes.size() ; i++) {
      IAxis yAxis = (IAxis)this.valueAxes.get(i) ;
      if (yAxis != null) {
        RectangleEdge edge = getValueAxisEdge(i) ;
        double fixedSpace = yAxis.getFixedAxisSpace() ;
        double axisWidth = 0 ;
        if (fixedSpace > 0) {
          axisWidth = fixedSpace ;
        } else {
          axisWidth = yAxis.calculateAxisSpace(g2, plotArea, clonePlotArea, edge) ;
        }

        axisSpace.add(axisWidth, edge) ;
        valueAxisWidthArr[i] = axisWidth ;
        edge.trim(clonePlotArea, axisWidth) ;
      }
    }

    RectangleInsets atLeastAxisOffset = getAxisOffset().ensureAtLeast(1) ; //ÿ߽Ϊ1
    atLeastAxisOffset.trim(clonePlotArea) ; //trim data area
    double axisOffsetWidth = atLeastAxisOffset.getLeft() + atLeastAxisOffset.getRight() ;
    double axisOffsetHeight = atLeastAxisOffset.getTop() + atLeastAxisOffset.getBottom() ;

    //Խ
    axisSpace.setTop(Math.min(plotArea.getHeight(), axisSpace.getTop())) ;
    axisSpace.setBottom(Math.min(plotArea.getHeight(), axisSpace.getBottom())) ;
    axisSpace.setLeft(Math.min(plotArea.getWidth(), axisSpace.getLeft())) ;
    axisSpace.setRight(Math.min(plotArea.getWidth(), axisSpace.getRight())) ;

    /////////////////////////// draw top axes ///////////////////////////////
    //draw top category axes
    double topSize = 0.0 ;
    for (int m = this.domainAxes.size() - 1 ; m >= 0 ; m--) {
      IAxis xAxis = (IAxis)this.domainAxes.get(m) ;
      if (xAxis != null) {
        RectangleEdge edge = getDomainAxisEdge(m) ;
        if (edge.equals(RectangleEdge.TOP)) {
          if (topSize + categoryAxisWidthArr[m] > plotArea.getHeight()) { //Խʾ
            break ;
          }

          Rectangle2D axisArea = new Rectangle2D.Double(plotArea.getX() + axisSpace.getLeft(),
                                                        plotArea.getY() + topSize,
                                                        clonePlotArea.getWidth() + axisOffsetWidth,
                                                        categoryAxisWidthArr[m]) ;
          AxisShapeCollection xAxisShapes = new AxisShapeCollection() ;
          xAxisShapes.setEdge(edge) ;
          xAxisShapes.setDrawArea(axisArea) ;

          topSize += categoryAxisWidthArr[m] ;
          xAxis.draw(g2, plotArea, axisArea, edge, xAxisShapes, false) ;
          shapeCollection.addXAxisShape(xAxisShapes) ;
        }
      }
    }
    //draw top value axes
    for (int n = this.valueAxes.size() - 1 ; n >= 0 ; n--) {
      IAxis yAxis = (IAxis)this.valueAxes.get(n) ;
      if (yAxis != null) {
        RectangleEdge edge = getValueAxisEdge(n) ;
        if (edge.equals(RectangleEdge.TOP)) {
          if (topSize + valueAxisWidthArr[n] > plotArea.getHeight()) { //Խʾ
            break ;
          }

          Rectangle2D axisArea = new Rectangle2D.Double(plotArea.getX() + axisSpace.getLeft(),
                                                        plotArea.getY() + topSize,
                                                        clonePlotArea.getWidth() + axisOffsetWidth,
                                                        valueAxisWidthArr[n]) ;
          AxisShapeCollection yAxisShapes = new AxisShapeCollection() ;
          yAxisShapes.setEdge(edge) ;
          yAxisShapes.setDrawArea(axisArea) ;

          topSize += valueAxisWidthArr[n] ;
          yAxis.draw(g2, plotArea, axisArea, edge, yAxisShapes, false) ;
          if (yAxis instanceof AbstractValueAxis) {
            int renderIndex = getRenderIndexForValueAxis(n) ;
            RenderShapeCollection renderShapes = getRenderShapeCollection(renderIndex) ;
            if (renderShapes != null) {
              ((AbstractValueAxis) yAxis).drawValueLine(g2, clonePlotArea, edge, renderShapes) ;
            }
          }
          shapeCollection.addYAxisShape(yAxisShapes) ;
        }
      }
    }

    /////////////////////////// draw bottom axes ///////////////////////////////
    //draw bottom category axes
    //һbottom 3d axisΪ2dʽ
    int firstBottom3DIndex = -1 ;
    int bottom3DAxisCount = 0 ;
    for (int i = 0 ; i < this.domainAxes.size() ; i++) {
      IAxis xAxis = (IAxis)this.domainAxes.get(i) ;
      if (xAxis != null && xAxis instanceof CategoryAxis3D) {
        RectangleEdge edge = getDomainAxisEdge(i) ;
        if (edge.equals(RectangleEdge.BOTTOM)) {
          if (firstBottom3DIndex < 0) {
            firstBottom3DIndex = i ;
          }
          bottom3DAxisCount++ ;
        }
      }
    }

    if (bottom3DAxisCount > 1) {
      for (int i = 0 ; i < this.domainAxes.size() ; i++) {
        IAxis xAxis = (IAxis)this.domainAxes.get(i) ;
        if (xAxis != null && xAxis instanceof CategoryAxis3D) {
          RectangleEdge edge = getDomainAxisEdge(i) ;
          if (edge.equals(RectangleEdge.BOTTOM)) {
            if (i != firstBottom3DIndex) {
              ((CategoryAxis3D) xAxis).setForce2D(true) ;
            }
          }
        }
      }
    }

    double bottomSize = 0.0 ;
    for (int m = this.domainAxes.size() - 1 ; m >= 0 ; m--) {
      IAxis xAxis = (IAxis)this.domainAxes.get(m) ;
      if (xAxis != null) {
        RectangleEdge edge = getDomainAxisEdge(m) ;
        if (edge.equals(RectangleEdge.BOTTOM)) {
          if (topSize + bottomSize + categoryAxisWidthArr[m] > plotArea.getHeight()) { //Խʾ
            break ;
          }

          bottomSize += categoryAxisWidthArr[m] ;
          Rectangle2D axisArea = new Rectangle2D.Double(plotArea.getX() + axisSpace.getLeft(),
                                                        plotArea.getMaxY() - bottomSize,
                                                        clonePlotArea.getWidth() + axisOffsetWidth,
                                                        categoryAxisWidthArr[m]) ;
          AxisShapeCollection xAxisShapes = new AxisShapeCollection() ;
          xAxisShapes.setEdge(edge) ;
          xAxisShapes.setDrawArea(axisArea) ;

          xAxis.draw(g2, plotArea, axisArea, edge, xAxisShapes, false) ;
          shapeCollection.addXAxisShape(xAxisShapes) ;
        }
      }
    }

    //draw bottom value axes
    //һleft 3d axisΪ2dʽ
    int firstBottomValue3DIndex = -1 ;
    int leftBottom3DAxisCount = 0 ;
    for (int i = 0 ; i < this.valueAxes.size() ; i++) {
      IAxis yAxis = (IAxis)this.valueAxes.get(i) ;
      if (yAxis != null && yAxis instanceof NumberValueAxis3D) {
        RectangleEdge edge = getValueAxisEdge(i) ;
        if (edge.equals(RectangleEdge.BOTTOM)) {
          if (firstBottomValue3DIndex < 0) {
            firstBottomValue3DIndex = i ;
          }
          leftBottom3DAxisCount++ ;
        }
      }
    }

    if (leftBottom3DAxisCount > 1) {
      for (int i = 0 ; i < this.valueAxes.size() ; i++) {
        IAxis yAxis = (IAxis)this.valueAxes.get(i) ;
        if (yAxis != null && yAxis instanceof NumberValueAxis3D) {
          RectangleEdge edge = getValueAxisEdge(i) ;
          if (edge.equals(RectangleEdge.BOTTOM)) {
            if (i != firstBottom3DIndex) {
              ((NumberValueAxis3D) yAxis).setForce2D(true) ;
            }
          }
        }
      }
    }

    for (int n = this.valueAxes.size() - 1 ; n >= 0 ; n--) {
      IAxis yAxis = (IAxis)this.valueAxes.get(n) ;
      if (yAxis != null) {
        RectangleEdge edge = getValueAxisEdge(n) ;
        if (edge.equals(RectangleEdge.BOTTOM)) {
          if (topSize + bottomSize + valueAxisWidthArr[n] > plotArea.getHeight()) { //Խʾ
            break ;
          }

          bottomSize += valueAxisWidthArr[n] ;
          Rectangle2D axisArea = new Rectangle2D.Double(plotArea.getX() + axisSpace.getLeft(),
                                                        plotArea.getMaxY() - bottomSize,
                                                        clonePlotArea.getWidth() + axisOffsetWidth,
                                                        valueAxisWidthArr[n]) ;
          AxisShapeCollection yAxisShapes = new AxisShapeCollection() ;
          yAxisShapes.setEdge(edge) ;
          yAxisShapes.setDrawArea(axisArea) ;

          yAxis.draw(g2, plotArea, axisArea, edge, yAxisShapes, false) ;
          if (yAxis instanceof AbstractValueAxis) {
            int renderIndex = getRenderIndexForValueAxis(n) ;
            RenderShapeCollection renderShapes = getRenderShapeCollection(renderIndex) ;
            if (renderShapes != null) {
              ((AbstractValueAxis) yAxis).drawValueLine(g2, clonePlotArea, edge, renderShapes) ;
            }
          }
          shapeCollection.addYAxisShape(yAxisShapes) ;
        }
      }
    }

    /////////////////////////// draw left axes ///////////////////////////////
    //draw left category axes
    //һbottom 3d axisΪ2dʽ
    int firstLeftCategory3DIndex = -1 ;
    int leftCategory3DAxisCount = 0 ;
    for (int i = 0 ; i < this.domainAxes.size() ; i++) {
      IAxis xAxis = (IAxis)this.domainAxes.get(i) ;
      if (xAxis != null && xAxis instanceof CategoryAxis3D) {
        RectangleEdge edge = getDomainAxisEdge(i) ;
        if (edge.equals(RectangleEdge.LEFT)) {
          if (firstLeftCategory3DIndex < 0) {
            firstLeftCategory3DIndex = i ;
          }
          leftCategory3DAxisCount++ ;
        }
      }
    }

    if (leftCategory3DAxisCount > 1) {
      for (int i = 0 ; i < this.domainAxes.size() ; i++) {
        IAxis xAxis = (IAxis)this.domainAxes.get(i) ;
        if (xAxis != null && xAxis instanceof CategoryAxis3D) {
          RectangleEdge edge = getDomainAxisEdge(i) ;
          if (edge.equals(RectangleEdge.LEFT)) {
            if (i != firstLeftCategory3DIndex) {
              ((CategoryAxis3D) xAxis).setForce2D(true) ;
            }
          }
        }
      }
    }

    double leftSize = 0.0 ;
    double rightSize = 0.0 ;
    boolean leftToRight = false ;
    for (int m = this.domainAxes.size() - 1 ; m >= 0 ; m--) {
      IAxis xAxis = (IAxis)this.domainAxes.get(m) ;
      if (xAxis != null) {
        RectangleEdge edge = getDomainAxisEdge(m) ;
        if (edge.equals(RectangleEdge.LEFT)) {
          if (xAxis instanceof IEffect3D && !leftToRight && categoryAxesCount <= 1) {
            if (rightSize + leftSize + categoryAxisWidthArr[m] > plotArea.getWidth()) { //Խʾ
              break ;
            }

            rightSize += categoryAxisWidthArr[m] ;
            IEffect3D effect = (IEffect3D) xAxis ;
            Rectangle2D axisArea = new Rectangle2D.
                                   Double(plotArea.getMaxX() - rightSize + effect.getXOffset(),
                                          plotArea.getY() + axisSpace.getTop() - effect.getYOffset(),
                                          categoryAxisWidthArr[m],
                                          clonePlotArea.getHeight() + axisOffsetHeight) ;
            AxisShapeCollection xAxisShapes = new AxisShapeCollection() ;
            xAxisShapes.setEdge(edge) ;
            xAxisShapes.setDrawArea(axisArea) ;

            xAxis.draw(g2, plotArea, axisArea, RectangleEdge.RIGHT, xAxisShapes, false) ;
            shapeCollection.addXAxisShape(xAxisShapes) ;
            leftToRight = true ;
          } else {
            if (leftSize + categoryAxisWidthArr[m] > plotArea.getWidth()) { //Խʾ
              break ;
            }

            Rectangle2D axisArea = new Rectangle2D.Double(plotArea.getX() + leftSize,
                                                          plotArea.getY() + axisSpace.getTop(),
                                                          categoryAxisWidthArr[m],
                                                          clonePlotArea.getHeight()
                                                          + axisOffsetHeight) ;
            AxisShapeCollection xAxisShapes = new AxisShapeCollection() ;
            xAxisShapes.setEdge(edge) ;
            xAxisShapes.setDrawArea(axisArea) ;

            leftSize += categoryAxisWidthArr[m] ;
            xAxis.draw(g2, plotArea, axisArea, edge, xAxisShapes, false) ;

            shapeCollection.addXAxisShape(xAxisShapes) ;
          }
        }
      }
    }

    //draw left value axes
    //һleft 3d axisΪ2dʽ
    int firstLeftValue3DIndex = -1 ;
    int leftValue3DAxisCount = 0 ;
    for (int i = 0 ; i < this.valueAxes.size() ; i++) {
      IAxis yAxis = (IAxis)this.valueAxes.get(i) ;
      if (yAxis != null && yAxis instanceof NumberValueAxis3D) {
        RectangleEdge edge = getValueAxisEdge(i) ;
        if (edge.equals(RectangleEdge.LEFT)) {
          if (firstLeftValue3DIndex < 0) {
            firstLeftValue3DIndex = i ;
          }
          leftValue3DAxisCount++ ;
        }
      }
    }

    if (leftValue3DAxisCount > 1) {
      for (int i = 0 ; i < this.valueAxes.size() ; i++) {
        IAxis yAxis = (IAxis)this.valueAxes.get(i) ;
        if (yAxis != null && yAxis instanceof NumberValueAxis3D) {
          RectangleEdge edge = getValueAxisEdge(i) ;
          if (edge.equals(RectangleEdge.LEFT)) {
            if (i != leftValue3DAxisCount) {
              ((NumberValueAxis3D) yAxis).setForce2D(true) ;
            }
          }
        }
      }
    }

    for (int n = this.valueAxes.size() - 1 ; n >= 0 ; n--) {
      IAxis yAxis = (IAxis)this.valueAxes.get(n) ;
      if (yAxis != null) {
        RectangleEdge edge = getValueAxisEdge(n) ;
        if (edge.equals(RectangleEdge.LEFT)) {
          if (leftSize + valueAxisWidthArr[n] > plotArea.getWidth()) { //Խʾ
            break ;
          }

          Rectangle2D axisArea = new Rectangle2D.Double(plotArea.getX() + leftSize,
                                                        plotArea.getY() + axisSpace.getTop(),
                                                        valueAxisWidthArr[n],
                                                        clonePlotArea.getHeight()
                                                        + axisOffsetHeight) ;
          AxisShapeCollection yAxisShapes = new AxisShapeCollection() ;
          yAxisShapes.setEdge(edge) ;
          yAxisShapes.setDrawArea(axisArea) ;

          leftSize += valueAxisWidthArr[n] ;
          yAxis.draw(g2, plotArea, axisArea, edge, yAxisShapes, false) ;
          if (yAxis instanceof AbstractValueAxis) {
            int renderIndex = getRenderIndexForValueAxis(n) ;
            RenderShapeCollection renderShapes = getRenderShapeCollection(renderIndex) ;
            if (renderShapes != null) {
              ((AbstractValueAxis) yAxis).drawValueLine(g2, clonePlotArea, edge, renderShapes) ;
            }
          }
          shapeCollection.addYAxisShape(yAxisShapes) ;
        }
      }
    }

    /////////////////////////// draw right axes ///////////////////////////////
    //draw right category axes
    for (int m = this.domainAxes.size() - 1 ; m >= 0 ; m--) {
      IAxis xAxis = (IAxis)this.domainAxes.get(m) ;
      if (xAxis != null) {
        RectangleEdge edge = getDomainAxisEdge(m) ;
        if (edge.equals(RectangleEdge.RIGHT)) {
          if (rightSize + leftSize + categoryAxisWidthArr[m] > plotArea.getWidth()) { //Խʾ
            break ;
          }

          rightSize += categoryAxisWidthArr[m] ;
          Rectangle2D axisArea = new Rectangle2D.Double(plotArea.getMaxX() - rightSize,
                                                        plotArea.getY() + axisSpace.getTop(),
                                                        categoryAxisWidthArr[m],
                                                        clonePlotArea.getHeight()
                                                        + axisOffsetHeight) ;
          AxisShapeCollection xAxisShapes = new AxisShapeCollection() ;
          xAxisShapes.setEdge(edge) ;
          xAxisShapes.setDrawArea(axisArea) ;

          xAxis.draw(g2, plotArea, axisArea, edge, xAxisShapes, false) ;
          shapeCollection.addXAxisShape(xAxisShapes) ;
        }
      }
    }
    //draw right value axes
    for (int n = this.valueAxes.size() - 1 ; n >= 0 ; n--) {
      IAxis yAxis = (IAxis)this.valueAxes.get(n) ;
      if (yAxis != null) {
        RectangleEdge edge = getValueAxisEdge(n) ;
        if (edge.equals(RectangleEdge.RIGHT)) {
          if (rightSize + leftSize + valueAxisWidthArr[n] > plotArea.getWidth()) { //Խʾ
            break ;
          }

          rightSize += valueAxisWidthArr[n] ;
          Rectangle2D axisArea = new Rectangle2D.Double(plotArea.getMaxX() - rightSize,
                                                        plotArea.getY() + axisSpace.getTop(),
                                                        valueAxisWidthArr[n],
                                                        clonePlotArea.getHeight()
                                                        + axisOffsetHeight) ;
          AxisShapeCollection yAxisShapes = new AxisShapeCollection() ;
          yAxisShapes.setEdge(edge) ;
          yAxisShapes.setDrawArea(axisArea) ;

          yAxis.draw(g2, plotArea, axisArea, edge, yAxisShapes, false) ;
          if (yAxis instanceof AbstractValueAxis) {
            int renderIndex = getRenderIndexForValueAxis(n) ;
            RenderShapeCollection renderShapes = getRenderShapeCollection(renderIndex) ;
            if (renderShapes != null) {
              ((AbstractValueAxis) yAxis).drawValueLine(g2, clonePlotArea, edge, renderShapes) ;
            }
          }
          shapeCollection.addYAxisShape(yAxisShapes) ;
        }
      }
    }

    return clonePlotArea ;
  }

  /**
   * <zh>ޱɫ.</zh>
   * <en>Fills the four quadrant area with the quadrant paint.</en>
   *
   * @param dataArea Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The chart data area (null not permitted).</en>
   * @param shapeCollection RenderShapeCollection <zh>plotͼμ (Ϊnull).</zh>
   * <en>Collects information about the plot (null not permitted).</en>
   */
  /**
   * <zh>ޱɫ.</zh>
   * <en>Fills the four quadrant area with the quadrant paint.</en>
   *
   * @param dataArea Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The chart data area (null not permitted).</en>
   * @param orientation ChartOrientation <zh>ͼƷ.</zh>
   * <en>The chart draw orientation.</en>
   * @param domainAxis AbstractValueAxis <zh>ص0domain axis.</zh>
   * <en>Returns the domain axis for the plot.</en>
   * @param valueAxis AbstractValueAxis <zh>ص0domain axis.</zh>
   * <en>Returns the domain axis for the plot.</en>
   * @param renderShapes RenderShapeCollection <zh>renderͼμ (Ϊnull).</zh>
   * <en>Collects information about the render (null not permitted).</en>
   */
  protected void drawQuadrantBackground(Rectangle2D dataArea, ChartOrientation orientation,
                                        AbstractValueAxis domainAxis, AbstractValueAxis valueAxis,
                                        RenderShapeCollection renderShapes) {
    //  1 | 0
    //  --+--
    //  2 | 3
    double x = this.quadrantOrigin.getX() ;
    double y = this.quadrantOrigin.getY() ;
    if (Double.isNaN(x) || Double.isNaN(y)) {
      return ;
    }
    double xPos = domainAxis.valueToJava2D(dataArea, getDomainAxisEdge(), x, false) ;
    double yPos = valueAxis.valueToJava2D(dataArea, getValueAxisEdge(), y, false) ;
    if (xPos == AbstractValueAxis.INVALID_COORDINATE_VALUE
        || yPos == AbstractValueAxis.INVALID_COORDINATE_VALUE) {
      return ;
    }

    double areaX = dataArea.getX() ;
    double areaY = dataArea.getY() ;
    double areaMaxX = dataArea.getMaxX() ;
    double areaMaxY = dataArea.getMaxY() ;
    Rectangle2D[] rectArr = new Rectangle2D[4] ;
    if (orientation.equals(ChartOrientation.VERTICAL)) {
      rectArr[0] = new Rectangle2D.Double(xPos, areaY, areaMaxX - xPos, yPos - areaY) ;
      rectArr[1] = new Rectangle2D.Double(areaX, areaY, xPos - areaX, yPos - areaY) ;
      rectArr[2] = new Rectangle2D.Double(areaX, yPos, xPos - areaX, areaMaxY - yPos) ;
      rectArr[3] = new Rectangle2D.Double(xPos, yPos, areaMaxX - xPos, areaMaxY - yPos) ;
    } else if (orientation.equals(ChartOrientation.HORIZONTAL)) {
      rectArr[0] = new Rectangle2D.Double(yPos, areaY, areaMaxX - yPos, xPos - areaY) ;
      rectArr[1] = new Rectangle2D.Double(areaX, areaY, yPos - areaX, xPos - areaY) ;
      rectArr[2] = new Rectangle2D.Double(areaX, xPos, yPos - areaX, areaMaxY - xPos) ;
      rectArr[3] = new Rectangle2D.Double(yPos, xPos, areaMaxX - yPos, areaMaxY - xPos) ;
    }

    for (int i = 0 ; i < 4 ; i++) {
      if (this.quadrantPaint != null && this.quadrantPaint[i] != null && rectArr[i] != null) {
        RectangleTextChartShape shape = new RectangleTextChartShape(rectArr[i], null) ;
        shape.setDrawBorder(false) ;
        shape.setFillPaint(this.quadrantPaint[i]) ;
        shape.setFillShape(true) ;

        renderShapes.addQuadrantBackgroundShape(shape) ;
      }
    }
  }

  /**
   * <zh>domain0ֵ.</zh>
   * <en>Draw zero line of domain axis.</en>
   *
   * @param dataArea Rectangle2D <zh>data (Ϊnull).</zh>
   * <en>The data area (null not permitted).</en>
   * @param plotShapes PlotShapeCollection <zh>plotͼμ (Ϊnull).</zh>
   * <en>Collects information about the plot (null not permitted).</en>
   */
  protected void drawZeroDomainBaseline(Rectangle2D dataArea, PlotShapeCollection plotShapes) {
    if (isDomainZeroBaselineVisible()) {
      AbstractValueAxis domainAxis = getDomainAxis() ;
      if (domainAxis != null && domainAxis instanceof NumberValueAxis) {
        RectangleEdge edge = getDomainAxisEdge() ;
        NumberValueAxis numberAxis = (NumberValueAxis) domainAxis ;
        double zeroPosition = numberAxis.valueToJava2D(dataArea, edge, 0, false) ;
        LineChartShape lineShape = null ;
        if (RectangleEdge.isTopOrBottom(edge)) {
          lineShape = new LineChartShape(zeroPosition, dataArea.getMinY(),
                                         zeroPosition, dataArea.getMaxY()) ;

        } else if (RectangleEdge.isLeftOrRight(edge)) {
          lineShape = new LineChartShape(dataArea.getMinX(), zeroPosition,
                                         dataArea.getMaxX(), zeroPosition) ;
        }

        plotShapes.setZeroDomainBaseline(lineShape) ;
      }
    }
  }

  /**
   * <zh>value0ֵ.</zh>
   * <en>Draw zero line of value axis.</en>
   *
   * @param dataArea Rectangle2D <zh>data (Ϊnull).</zh>
   * <en>The data area (null not permitted).</en>
   * @param plotShapes PlotShapeCollection <zh>plotͼμ (Ϊnull).</zh>
   * <en>Collects information about the plot (null not permitted).</en>
   */
  protected void drawZeroValueBaseline(Rectangle2D dataArea, PlotShapeCollection plotShapes) {
    if (isValueZeroBaselineVisible()) {
      AbstractValueAxis valueAxis = getValueAxis() ;
      if (valueAxis != null && valueAxis instanceof NumberValueAxis) {
        RectangleEdge edge = getValueAxisEdge() ;
        NumberValueAxis numberAxis = (NumberValueAxis) valueAxis ;
        double zeroPosition = numberAxis.valueToJava2D(dataArea, edge, 0, false) ;
        LineChartShape lineShape = null ;
        if (RectangleEdge.isTopOrBottom(edge)) {
          lineShape = new LineChartShape(zeroPosition, dataArea.getMinY(),
                                         zeroPosition, dataArea.getMaxY()) ;

        } else if (RectangleEdge.isLeftOrRight(edge)) {
          lineShape = new LineChartShape(dataArea.getMinX(), zeroPosition,
                                         dataArea.getMaxX(), zeroPosition) ;
        }

        plotShapes.setZeroValueBaseline(lineShape) ;
      }
    }
  }

  /**
   * <zh>ص0render.</zh>
   * <en>Returns a reference to the render for the plot.</en>
   *
   * @return AbstractXYRender <zh>0render</zh>
   * <en>The render.</en>
   */
  public AbstractXYRender getRender() {
    return getRender(0) ;
  }

  /**
   * <zh>ֵȡrender.</zh>
   * <en>Returns the render at the given index.</en>
   *
   * @param index <zh>ֵ,ڵ0.</zh>
   * <en>The render index (zero-based).</en>
   *
   * @return AbstractXYRender <zh>render(Ϊnull).</zh>
   * <en>The render (possibly null).</en>
   */
  public AbstractXYRender getRender(int index) {
    if (index < 0) {
      throw new IllegalArgumentException("Must be 'index'>=0.") ;
    }

    AbstractXYRender result = null ;
    if (this.renders.size() > index) {
      result = (AbstractXYRender)this.renders.get(index) ;
    }

    return result ;
  }

  /**
   * <zh>render,˼ϲ޸.</zh>
   * <en>Returns the collection of renders,and this collection is unmodifiable.</en>
   *
   * @return Collection <zh>render.</zh>
   * <en>The collection of renders.</en>
   */
  public Collection getRenders() {
    return Collections.unmodifiableCollection(this.renders) ;
  }

  /**
   * <zh>õ0render.</zh>
   * <en>Sets the render at index 0 (sometimes referred to as the "primary" render) and sends
   * a {@link com.ijchart.xychart.event.PlotChangeEvent} to all registered listeners.</en>
   *
   * @param render AbstractXYRender <zh>render (Ϊnull).</zh>
   * <en>The render (null not permitted.</en>
   */
  public void setRender(AbstractXYRender render) {
    setRender(0, render, true) ;
  }

  /**
   * <zh>õ0render,Ҫм
   * {@link com.ijchart.xychart.event.PlotChangeEvent}¼.</zh>
   * <en>Sets the render at index 0 (sometimes referred to as the "primary" render) and,
   * if requested, sends a {@link com.ijchart.xychart.event.PlotChangeEvent}
   *  to all registered listeners.</en>
   *
   * @param render  AbstractXYRender <zh>render (Ϊnull).</zh>
   * <en>The render (null not permitted).</en>
   * @param notify  boolean <zh>Ƿ֪ͨ.</zh>
   * <en>Whether or not notify listeners.</en>
   */
  public void setRender(AbstractXYRender render, boolean notify) {
    setRender(0, render, notify) ;
  }

  /**
   * <zh>趨ֵָrender,м
   * {@link com.ijchart.xychart.event.PlotChangeEvent}¼.</zh>
   * <en>Sets the render at the specified index and sends a
   * {@link com.ijchart.xychart.event.PlotChangeEvent} to all registered listeners.</en>
   *
   * @param index int <zh>ֵ,ڵ0.</zh>
   * <en>The index (zero-based).</en>
   * @param render AbstractXYRender <zh>render (Ϊnull).</zh>
   * <en>The render (null not permitted).</en>
   */
  public void setRender(int index, AbstractXYRender render) {
    setRender(index, render, true) ;
  }

  /**
   * <zh>趨ֵָrender,м
   * {@link com.ijchart.xychart.event.PlotChangeEvent}¼.</zh>
   * <en>Sets a render.  A {@link com.ijchart.xychart.event.PlotChangeEvent}
   *  is sent to all registered listeners.</en>
   *
   * @param index int <zh>ֵ,ڵ0.</zh>
   * <en>The dataset index (zero-based).</en>
   * @param render AbstractXYRender <zh>render (Ϊnull).</zh>
   * <en>The render (null not permitted).</en>
   * @param notify  boolean <zh>Ƿ֪ͨ.</zh>
   * <en>Whether or not notify listeners.</en>
   */
  public void setRender(int index, AbstractXYRender render, boolean notify) {
    if (render == null) {
      throw new IllegalArgumentException("Null 'render' argument.") ;
    }

    // stop listening to the existing render...
    if (index < this.renders.size()) {
      AbstractXYRender existing = (AbstractXYRender)this.renders.get(index) ;
      if (existing != null) {
        existing.removeChangeListener(this) ;
      }
    }

    // register the new render...
    setListObject(renders, index, render) ;
    render.addChangeListener(this) ;

    if (notify) {
      firePlotChanged() ;
    }
  }

  /**
   * <zh>render,м{@link com.ijchart.xychart.event.PlotChangeEvent}¼.</zh>
   * <en>Sets the renders for this plot and sends a {@link com.ijchart.xychart.event.PlotChangeEvent}
   * to all registered listeners.</en>
   *
   * @param renders AbstractXYRender[] <zh>render (Ϊnull).</zh>
   * <en>The renders (null not permitted).</en>
   * @param startIndex int <zh>Ŀʼֵ,startIndex֮ǰĶ󲻻ᱻ,ڵ0.</zh>
   * <en>The beginning of the index value, startIndex before the target will not be covered.
   * must be greater than or equal to 0.</en>
   */
  public void setRenders(AbstractXYRender[] renders, int startIndex) {
    if (renders == null) {
      throw new IllegalArgumentException("Null 'renders' argument.") ;
    }
    if (startIndex < 0) {
      throw new IllegalArgumentException("Must be 'startIndex'>=0.") ;
    }

    if (renders.length > 0) {
      for (int i = 0 ; i < renders.length ; i++) {
        setRender(startIndex + i, renders[i], false) ;
      }
      firePlotChanged() ;
    }
  }

  /**
   * <zh>ָrender,ûҵ򷵻-1.</zh>
   * <en>Returns the index of the specified render, or -1 if the
   * render is not assigned to this plot.</en>
   *
   * @param render AbstractXYRender <zh>render (Ϊnull).</zh>
   * <en>The render (null permitted).</en>
   *
   * @return int <zh>ֵ.</zh>
   * <en>The render index.</en>
   */
  public int getRenderIndex(AbstractXYRender render) {
    return this.renders.indexOf(render) ;
  }

  /**
   * <zh>plotַ.</zh>
   * <en>Returns a short string describing the plot type.</en>
   *
   * @return <zh>plot.</zh>
   * <en>The plot type.</en>
   */
  public PlotType getPlotType() {
    return PlotType.PLOT_TYPE_XYPLOT ;
  }

  /**
   * <zh>Ƶ0category᷽.</zh>
   * <en>Draws the first category axis gridlines for the plot.</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(ӡĻ) (Ϊnull).</zh>
   * <en>The graphics device (null not permitted).</en>
   * @param dataArea Rectangle2D <zh>ͼdata() (Ϊnull).</zh>
   * <en>The area within which the data should be drawn (null not permitted).</en>
   * @param renderShapes RenderShapeCollection <zh>renderͼμ (Ϊnull).</zh>
   * <en>Collects information about the render (null not permitted).</en>
   */
  private void drawDomainGridlines(Graphics2D g2, Rectangle2D dataArea,
                                   RenderShapeCollection renderShapes) {
    if (isDomainGridlineVisible() && this.domainAxes.size() > 0) {
      RectangleEdge axisEdge = getDomainAxisEdge() ;
      Stroke gridStroke = getDomainGridlineStroke() ;
      Paint gridPaint = getDomainGridlinePaint() ;

      AbstractAxis axis = (AbstractAxis)this.domainAxes.get(0) ;
      if (axis != null) {
        axis.drawGridLine(g2, dataArea, axisEdge, renderShapes, gridStroke, gridPaint) ;
      }
    }
  }

  /**
   * <zh>hashcode.</zh>
   * <en>Returns a hash code for this object.</en>
   *
   * @return <zh>hashcode.</zh>
   * <en>A hash code.</en>
   */
  public int hashCode() {
    int result = super.hashCode() ;
    result = 29 * result + HashcodeUtil.hashCode(this.domainAxisLocations) ;
    result = 29 * result + HashcodeUtil.hashCode(this.renders) ;
    result = 29 * result + HashcodeUtil.hashCode(this.quadrantOrigin) ;
    result = 29 * result + HashcodeUtil.hashCode(this.quadrantPaint) ;

    return result ;
  }

  /**
   * Tests this plot for equality with another object.
   *
   * @param obj  the object (null permitted).
   *
   * @return true or false.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof XYPlot)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    XYPlot that = (XYPlot) obj ;
    if (!EqualsUtil.equals(this.renders, that.renders)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.quadrantOrigin, that.quadrantOrigin)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.domainAxisLocations, that.domainAxisLocations)) {
      return false ;
    }
    if (this.domainZeroBaselineVisible != that.domainZeroBaselineVisible) {
      return false ;
    }
    if (!EqualsUtil.equals(this.domainZeroBaselinePaint, that.domainZeroBaselinePaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.domainZeroBaselineStroke, that.domainZeroBaselineStroke)) {
      return false ;
    }
    if (this.valueZeroBaselineVisible != that.valueZeroBaselineVisible) {
      return false ;
    }
    if (!EqualsUtil.equals(this.valueZeroBaselinePaint, that.valueZeroBaselinePaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.valueZeroBaselineStroke, that.valueZeroBaselineStroke)) {
      return false ;
    }
    if (!this.quadrantOrigin.equals(that.quadrantOrigin)) {
      return false ;
    }
    for (int i = 0 ; i < 4 ; i++) {
      if (!EqualsUtil.equals(this.quadrantPaint[i], that.quadrantPaint[i])) {
        return false ;
      }
    }

    return true ;
  }

  /**
   * Returns a clone of the plot.
   *
   * @return A clone.
   *
   * @throws CloneNotSupportedException  this can occur if some component of
   *         the plot cannot be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    XYPlot clone = (XYPlot)super.clone() ;

    clone.domainAxes = new ArrayList() ;
    for (int i = 0 ; i < this.domainAxes.size() ; i++) {
      AbstractValueAxis xAxis = (AbstractValueAxis)this.domainAxes.get(i) ;
      if (xAxis != null) {
        xAxis.removeAllChangeListener() ;
        AbstractValueAxis clonedAxis = (AbstractValueAxis) xAxis.clone() ;
        clone.setDomainAxis(i, clonedAxis) ;
      }
    }

    clone.valueAxes = new ArrayList() ;
    for (int i = 0 ; i < this.valueAxes.size() ; i++) {
      AbstractValueAxis yAxis = (AbstractValueAxis)this.valueAxes.get(i) ;
      if (yAxis != null) {
        yAxis.removeAllChangeListener() ;
        AbstractValueAxis clonedAxis = (AbstractValueAxis) yAxis.clone() ;
        clone.setValueAxis(i, clonedAxis) ;
      }
    }

    clone.quadrantPaint = (Paint[])this.quadrantPaint.clone() ;
    return clone ;

  }

  /**
   * <zh>toString().</zh>
   * <en>Override toString method.</en>
   *
   * @return String
   */
  public String toString() {
    return "XYPlot" ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the output stream.
   *
   * @throws IOException  if there is an I/O error.
   */
  private void writeObject(ObjectOutputStream stream) throws IOException {
    stream.defaultWriteObject() ;
    SerialUtil.writeStroke(stream, this.valueZeroBaselineStroke) ;
    SerialUtil.writePaint(stream, this.valueZeroBaselinePaint) ;
    for (int i = 0 ; i < 4 ; i++) {
      SerialUtil.writePaint(stream, this.quadrantPaint[i]) ;
    }
    SerialUtil.writeStroke(stream, this.domainZeroBaselineStroke) ;
    SerialUtil.writePaint(stream, this.domainZeroBaselinePaint) ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the input stream.
   *
   * @throws IOException  if there is an I/O error.
   * @throws ClassNotFoundException  if there is a classpath problem.
   */
  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject() ;
    this.valueZeroBaselineStroke = SerialUtil.readStroke(stream) ;
    this.valueZeroBaselinePaint = SerialUtil.readPaint(stream) ;
    this.quadrantPaint = new Paint[4] ;
    for (int i = 0 ; i < 4 ; i++) {
      this.quadrantPaint[i] = SerialUtil.readPaint(stream) ;
    }

    this.domainZeroBaselineStroke = SerialUtil.readStroke(stream) ;
    this.domainZeroBaselinePaint = SerialUtil.readPaint(stream) ;
  }

}
